/* Module anoncookies.c - Rev 1.11 - 26-Oct-1998 */

/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 *   COPYRIGHT AND USAGE CONDITIONS OF THIS MODULE
 *
 *   This file is (C) 1998 Vittorio Bertola <v.bertola@bertola.eu.org>.
 *   It can be freely used for no-profit, non commercial purposes. Usage
 *   in commercial or profit sites (e.g. ISPs, sites with banners, corporate
 *   sites) is forbidden without an explicit authorization from the author.
 *   Contact the author to discuss the matter :-)
 *
 *   THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY AND ITS AUTHOR
 *   WILL NOT BE HELD RESPONSIBLE FOR ANY DIRECT OR INDIRECT DAMAGE RELATED
 *   TO ITS USE. CAN YOU READ THE ABOVE DISCLAIMER FOR THE APACHE SERVER?
 *   THE SAME APPLIES TO ME AND THIS MODULE.
 *
 */

/*
 *                            --> USER'S GUIDE <--
 *
 *
 *   INTRODUCTION
 *
 *   This module allows to limit the maximum number of contemporary visitors
 *   in (sections of) a WWW site, mostly like the "maximum number of
 *   anonymous users" in a FTP site. However, it is much more flexible than
 *   the anonymous users limit in a FTP site: it can be set per directory,
 *   and different classes of users, with different access properties, can
 *   be defined.
 *   To be able to access limited directories, the user must previously have
 *   received an "anonymous cookie" of the right class(es); this module also
 *   cares about cookie distribution and expiring. Each class of cookies is
 *   marked with a maximum number of cookies in the system (when all of them
 *   have been distributed, no more cookies of that class will be available
 *   until someone gives his cookie back), a maximum idletime (i.e. time in
 *   which the user is not downloading anything) and a maximum number of 
 *   contemporary downloads. Then, per each directory in the system you can
 *   define whether anonymous cookies are required and which classes of
 *   cookies let the visitor in.
 *   To overcome the limits of some clients, it is possible to turn on for
 *   specified hosts or user-agent an alternative mode: the "direct" mode,
 *   that allows the users to get a ticket that lies only on the server
 *   side (so it works even if the user agent does not support cookies),
 *   with the disadvantage that the user is identified only by his IP
 *   address (so this won't work well for users with proxies) and that
 *   he cannot be granted any idletime: the ticket gets released as soon
 *   as he stops downloading files.
 *   In the end, it is also possible to exclude specified hosts/user agents
 *   from any access control based on anonymous cookies, and let them in
 *   without limits.
 *   As a side feature, it is also possible to limit the ability of other
 *   sites to link your pages/files: you can tell the server to reject all
 *   requests not coming from a link on a page on your site.
 *   This module uses cookies to give access to users: if the user has a 
 *   browser that doesn't support cookies, or if he disabled cookies, he
 *   will not be able to get in the protected directories.
 *
 *
 *
 *   COMPILING THE MODULE
 *
 *   1) Enter the Apache sources dir.
 *   2) Edit the 'Configuration' file and add the following line just after
 *      the "example_module" line, near the bottom of the file:
 *      
 *      Module anoncookies_module      mod_anoncookies.o
 *
 *   3) Run the 'Configure' script (command:   ./Configure   )
 *   4) Make the server (command:     make     )
 *   5) Take the resulting 'httpd' file and substitute your old 'httpd'
 *
 *
 *
 *   PER-SERVER CONFIGURATION
 *
 *   SetAnonCookieClass name [max]
 *
 *   Defines a class of anoncookies named 'name' and a (per-server) limit
 *   of max different cookies of that class that can coexist in the system.
 *   If max is omitted, infinite cookies of class 'name' will be given.
 *   Reserved names (that cannot be used as a class name) are 'no' and 'all'.
 *   IMPORTANT: Cookie classes must be defined with this command prior to
 *   be used in any other command.
 *
 *
 *   SetAnonCookieIdletime timeout (class1 [class2...] | all)
 *
 *   Sets the expiration timeout for cookies of the specified class(es); a
 *   cookie will be considered "dead" and freed when at least timeout
 *   seconds have passed since it stopped downloading files/pages. If 'all'
 *   is given as class name, the timeout is set for all classes already
 *   defined. At least one class name, or 'all', must be given.
 *
 *
 *   SetAnonCookieMaxFiles nfiles (class1 [class2...] | all)
 *
 *   Sets the maximum number of contemporary downloads that can be made
 *   by a user having a (single) cookie of the specified class(es). (See above
 *   for the second argument.)
 *
 *
 *   SetAnonCookieName name
 *   SetAnonCookiePath path
 *   SetAnonCookieDomain domain
 *
 *   Set the name, path and domain of the cookies sent to the users. (See
 *   cookies specification for that :-)  (They default to 'ANONCOOKIE',
 *   '/' and nothing (i.e. the server name).)
 *
 *
 *   SetAnonCookieDir dir
 *
 *   Sets the directory in which cookies will be stored (subdirs for each
 *   class will be created). The directory must be rwx by user nobody.
 *   Create it before running the server, and use this command before any
 *   class definition, at the very beginning of the anoncookies commands.
 *
 *
 *   SetAnonCookieOkDoc uri (class1 [class2...] | all)
 *
 *   Sets the URI (starting with /, relative to document root) that must
 *   be sent to the user when he gets a cookie of the specified class(es)
 *   (or 'all' to mean all classes, see SetAnonCookieIdletime). If this
 *   command is not given, a "standard" message is sent.
 *
 *
 *   SetAnonCookieErrorDoc uri (error1 [error2...] | all)
 *
 *   Sets the URI (starting with /, relative to document root) that must
 *   be sent to the user when the specified type(s) of error(s) arise (or
 *   'all' to mean all types of errors). The conventional names of the errors
 *   are defined below, in the XXXX_ITEM #defines, along with a short
 *   explanation of their meaning. If no error document is specified for an
 *   error, a standard message is sent.
 *  
 *
 *
 *   PER-DIRECTORY CONFIGURATION
 *
 *   RequireAnonCookies class1 [class2...]| no
 *
 *   Access in the specified directory is allowed only for users having
 *   a cookie of the specified class(es). If 'no' is specified, no cookies
 *   are required.
 *
 *
 *   AnonCookiesDirect always *
 *   AnonCookiesDirect never *
 *   AnonCookiesDirect with user-agent1 [user-agent2...]
 *   AnonCookiesDirect from host1 [host2...]
 *   AnonCookiesDirectClass class
 *
 *   Enables "direct mode". In direct mode, when an user without a valid
 *   cookie tries to get a file, he won't get an error message; instead, the
 *   server will try to grant him a cookie of the class specified by the
 *   'AnonCookiesDirectClass' command. The cookie is not sent to the user,
 *   but is stored on the server and tied to his IP address; any subsequent
 *   request from that IP will be treated just as the user had sent the
 *   cookie related to his address. So all normal limits on maximum files
 *   and idle time will apply, but the user-agent doesn't need to support
 *   cookies! The disadvantage is that there is not a real identification
 *   for the user: e.g. users behind a proxy will seem to be as one,
 *   the same will happen for multiple users of an Unix machine, and so on.
 *   However, it is useful to enable direct access for user-agents that
 *   don't (easily) support cookies (e.g. Lynx or GetRight).
 *   'always *' and 'never *' enable/disable direct mode for all requests;
 *   'with' enables direct mode for specified user-agents (identified
 *   by the codename); 'from' enables direct mode for specified hosts/IP
 *   addresses. Examples:
 *
 *     AnonCookiesDirect with GetRight/ Mozilla/1.0 Mozilla/1.1
 *     AnonCookiesDirect from 130.192. .polito.it
 *
 *   If no DirectClass is specified, the direct mode will always be
 *   disabled; you have to use the AnonCookiesDirectClass command _before_
 *   any AnonCookiesDirect command in each directory. 
 *   The user will still get an error message if no more cookies of that
 *   class are available, or if he tries to overcome class limits.
 *
 *
 *   AnonCookiesBypass with user-agent1 [user-agent2...]
 *   AnonCookiesBypass from host1 [host2...]
 *
 *   Excludes from any anonymous cookie limitation the specified user agents
 *   or hosts. Access is transparent and always allowed, and there is no
 *   limit on the maximum number of contemporary downloads.
 *
 *
 *   RejectExternalRefs on|off|explicit
 *
 *   If set to on, all requests coming from an external "referer" (i.e. links
 *   in pages on other servers), or not having any referer indication (i.e.
 *   manually entered by the user) will be refused. Only requests having our
 *   "canonical hostname" as referer will be accepted. You might want to
 *   set this to make visitors get cookies and files only from _your_ pages
 *   (reading your warnings and messages, etc.).
 *   The 'explicit' setting means that requests having an external referer
 *   will be refused, but those having no referer will be accepted: this
 *   because some old browsers don't support the "Referer" header and will
 *   be cut off from the site. However, this will allow users to get in
 *   by manually typing the URL, so... you decide.
 *   Defaults to 'off'.
 *   (This feature is completely independant from anonymous cookies... you
 *   can use it on non-restricted directories.)
 *
 *
 *
 *   HANDLERS
 *
 *   get-anoncookie
 *
 *   Allows you to distribute anonymous cookies to the users. It takes the
 *   last path element as the name of the requested class, and creates (if
 *   possible) a new cookie, that is sent to the user.
 *   
 *
 *
 *   TYPICAL CONFIGURATION AND USAGE
 *
 *   Here's what you usually put in access.conf (follow this order!):
 *
 *   1) Define a Location to distribute the anonymous cookies:
 *
 *      <Location /getcookies>
 *      SetHandler get-anoncookie
 *      </Location>
 *
 *      Now addresses of the form
 *
 *      http://www.your.site.net/getcookies/class
 *      or
 *      http://www.your.site.net/getcookies/class/
 *
 *      will generate the request of a cookie of class 'class', that will
 *      be sent to the user-agent if available.
 *      You can use the 'order', 'allow', 'deny' statements to limit the
 *      access to the location; you can have different limits by using
 *      different <Location>s, e.g.
 *
 *      <Location /getcookies/class1>
 *      SetHandler get-anoncookie
 *      order deny,allow
 *      deny from all
 *      allow from my.pc.net
 *      </Location>
 *      <Location /getcookies/class2>
 *      ...
 *
 *
 *   2) Define the cookies dir (with SetAnonCookieDir) and all the classes
 *      you are going to use (with SetAnonCookieClass):
 *
 *      SetAnonCookieDir /httpd/acookies
 *      SetAnonCookieClass class1
 *      SetAnonCookieClass class2 10
 *      ...
 *
 *      The AnonCookieDir must exist and have rwx privileges for user
 *      'nobody' (or the user the server runs as, defined in the 'User'
 *      configuration command in httpd.conf). (Create it, chown it to
 *      that user and chmod it to 0700.)
 *
 *   3) Define class features (e.g. idletimes, maxfiles, error and ok 
 *      documents...)
 *
 *   4) Only _after_ that, you can use the per-directory commands in
 *      the following <Directory> and <Location> sections: e.g.
 *
 *      <Directory /www/limited>
 *      RequireAnonCookies class1 class3
 *      (all other dir options, e.g. order/allow/deny, IndexOptions...)
 *      </Directory>
 *
 *
 *
 *   NOTES
 *
 *   -->Sub-requests, includes, dir listings<-- : the access restrictions
 *   are evaluated for the URI requested by the user. If the user requests
 *   a document outside of the protected dirs, and that document includes
 *   or redirects itself to a file in the protected dirs, *no cookie will
 *   be requested*. This allows you to bypass the protections, if you want
 *   to let users "have a look inside" before getting a cookie. However,
 *   if you think that this is a problem, let's talk about it :-)
 * 
 *   -->Limiting access to HTML files<-- : this module was conceived to 
 *   limit access to "simil-FTP" WWW sites: lists of files that can be
 *   downloaded. You can use it to block access to directories in which
 *   normal HTML pages reside, but you will have to do it cleverly. For
 *   example, if your cookie allows to download one file at a time, and
 *   your HTML page uses an image that is also in the protected section,
 *   the browser will try to download both the page and the image at the
 *   same time and will possibly get back an error!! So, the good way to
 *   do this is to put all HTML pages in the limited dir, and all images
 *   in a non-limited dir. The same applies to frames: limit the access,
 *   let's say, to right frame pages, and put left frame pages in a non
 *   protected dir. However, you can solve this problem also by rising
 *   the Files limit for the cookie class.
 *
 *   -->Inheritance of per-dir configs<-- : there is no inheritance.
 *   The configuration for the deepest dir overrides the one for the
 *   parent one.
 *
 *   -->Manual access to cookies dir<-- : you can safely delete cookies
 *   (obviously, the owner of the cookie won't like it :-). You can even add
 *   them, but be careful to add cookies with a correct name: the format
 *   is
 *
 *   <cookiename>.<expire>.<nfiles>
 * 
 *   where <cookiename> is an unique string of ANONCOOKIE_NAMELEN alphabetic
 *   chars (case sensitive), <expire> is the time_t expiration moment
 *   (-1 for no expire), <nfiles> is the number of files being downloaded
 *   (when you create a cookie, set it to 0). You can even define non-expiring
 *   cookies in classes having a finite idletime (or, better, make an existing
 *   cookie non-expiring by setting to -1 its <expire> - however, your browser
 *   will possibly delete the cookie when closing itself).
 *   However, don't mess with directories or you'll get some ANCK_DIR_ERRORs;
 *   in the case, kill and restart the server to get things right.
 * 
 *
 *
 *   HISTORY
 *
 *   v1.0 - 6/Oct/98   First released version.
 *   v1.1 - 23/Oct/98  Fixed some bugs: added resetting cookies in class dir
 *                     on server start, no-cache Pragma directive for proxies,
 *                     no expiring on client side. Added "direct mode" and
 *                     related configuration commands.
 *   v1.11- 26/Oct/98  Added "bypass" mode.
 *                     
 */ 

/* Note per il programmatore: in alloc.h ci sono i prototipi di molte funzioni utili...
   In http_config.h ci sono altre strutture e define interessanti. */



#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"

#include <stdio.h>
#include <errno.h>

#define ANONCOOKIE_NAMELEN 6          /* Length of the "given name" of the cookies */
#define ANONCOOKIES_ID "ANONCOOKIE"   /* Default AnonCookieName */
#define ANONCOOKIES_PATH "/"          /* Default AnonCookiePath */
#define ANONCOOKIES_DEFAULT_DIR "/httpd/acookies/"   /* Default cookies dir, must start and end with / */
#define ANCK_DIRECT_SUBDIR ".direct/"  /* Subdir for direct-mode cookies - must start with . */

#define ANCK_ERROR_HANDLER_NAME "senderror-anoncookies"
#define THE_PAST "Tue, 29-Sep-98 07:40:00 GMT"
#define THE_FUTURE "Thu, 31-Dec-20 23:59:00 GMT"

#define ANCK_INFINITE_COOKIES -1
#define ANCK_INFINITE_TIMEOUT -1
#define ANCK_DEFAULT_TIMEOUT -1
#define ANCK_INFINITE_MAXFILES -1
#define ANCK_DEFAULT_MAXFILES 1

#define MAX_SLEN 50          /* s deve contenere expiretime + nfiles in cifre */

#define N_ERRORS 11

#define NEED_ANONCOOKIE 0
#define WRONG_ANONCOOKIE 1
#define SPURIOUS_ANONCOOKIE 2
#define UNKNOWN_ANONCOOKIE 3
#define DEAD_ANONCOOKIE 4
#define TOO_MANY_ANONCOOKIES 5
#define MAXFILES_REACHED 6
#define DIDNT_UPDATE_ANONCOOKIE 7
#define MEMORY_ERROR 8
#define ANCK_DIR_ERROR 9
#define REJECT_XREFS 10

#define NEED_ANONCOOKIE_ITEM "Need"            /* The user hasn't got any anonymous cookie */
#define WRONG_ANONCOOKIE_ITEM "Wrong"          /* The user has an anoncookie of the wrong class (not the ones required for the dir) */
#define SPURIOUS_ANONCOOKIE_ITEM "Strange"     /* The user has an anoncookie of the right class but invalid (usually old, or even faked) */
#define UNKNOWN_ANONCOOKIE_ITEM "Unknown"      /* The user has an anoncookie of an unknown class */
#define DEAD_ANONCOOKIE_ITEM "Dead"            /* The user has an expired anoncookie (note: if the cookie is "recycled" before he tries to use it, the error is "Strange") */
#define TOO_MANY_ANONCOOKIES_ITEM "TooMany"    /* There are no free cookies of that class to be sent to the user */
#define MAXFILES_REACHED_ITEM "MaxFiles"       /* The user is already downloading MaxFiles files */
#define DIDNT_UPDATE_ANONCOOKIE_ITEM "NoUpdate"  /* The system couldn't update the cookie file - internal error, or another concurring thread recycled the cookie in the meanwhile */
#define MEMORY_ERROR_ITEM "Memory"             /* We couldn't allocate something */
#define ANCK_DIR_ERROR_ITEM "Dir"              /* We couldn't read/write in the cookies dir */
#define REJECT_XREFS_ITEM "XRef"               /* Someone tried to access with an external referer, and we don't like it */



typedef struct anoncookie_data {
    char *name;        /* Nome del cookie (lunghezza fissa per normali e var. per diretti) */
    int nfiles;        /* Numero di file in scaricamento */
    time_t expiretime; /* Istante (in epoch-s) in cui il cookie muore */
    char *filename;    /* (Vecchio) filename del cookie */
    char *dir;         /* Directory del cookie */
    pool *pool;        /* Pool in cui e' stato memorizzato */
} anoncookie_data;

typedef struct anoncookie_class {
    char *class;           /* Nome della classe */
    char *dir;             /* Directory in cui sono salvati i suoi cookie */
    char *directdir;       /* Directory per i cookie diretti */
    int maxcookies;        /* Numero massimo di cookie di tale classe */
    int classtimeout;      /* Idle timeout dei cookie della classe */
    int maxfiles;          /* Numero massimo di file contemporanei */
    char *ok_doc;          /* URL del documento da spedire quando assegno il cookie */
} anoncookie_class;

typedef struct anoncookie_reqconfig {
    anoncookie_data *cookie;
    anoncookie_class *class;
    int error;
} anoncookie_reqconfig;    

typedef struct anoncookie_srvconfig {
    array_header *classes;   /* Vettore di anoncookie_class */
    char *setcookie_name;
    int setcookie_namelen;
    char *setcookie_path;
    char *setcookie_domain;
    char *cookies_dir;
    array_header *error_docs;
} anoncookie_srvconfig;

typedef struct anoncookie_dirconfig {
    array_header *req_classes;
    array_header *direct_from;
    array_header *direct_with;
    array_header *bypass_from;
    array_header *bypass_with;
    anoncookie_class *direct_class;
    int reject_xrefs;
    int direct_always;
} anoncookie_dirconfig;    
    
/* La per srv config e' un anoncookie_srvconfig */
/* La per dir config e' un vettore di char* */    

/** DEBUG **/
/* #define DEBUG */

#define LOG "/etc/anoncookie.log"
#ifdef DEBUG
void aclog(char *);
static pool *anoncookies_pool = NULL;   /* Pool privata del modulo */
#endif

module anoncookies_module;
static int anoncookies_error_handler(request_rec *r);



#ifdef DEBUG
static void anoncookies_init(server_rec *s, pool *p) {

    if (anoncookies_pool == NULL) 
        anoncookies_pool = make_sub_pool(NULL);   /* Crea la pool privata di modulo */

}    


void aclog(char *s) {
    
    FILE *lf;
    
    if (anoncookies_pool == NULL)
        anoncookies_init(NULL, NULL);    
    
    lf = pfopen(anoncookies_pool, LOG, "a");
    if (lf != NULL) {
        if (s==NULL)
            fprintf(lf, "Argomento NULL al log.\n");
        else
            fprintf(lf, "%s\n", s);
        pfclose(anoncookies_pool, lf);
    }        

}
#endif



/*** CREAZIONE DELLE CONFIGURAZIONI ***/

static void *anoncookies_dircfg_create(pool *p, char *dirspec) {

    anoncookie_dirconfig *cfg;

    cfg = (anoncookie_dirconfig *) palloc(p, sizeof(anoncookie_dirconfig));
    if (cfg) {
        cfg->req_classes = make_array(p, 1, sizeof(char*));
        cfg->direct_from = make_array(p, 1, sizeof(char*));
        cfg->direct_with = make_array(p, 1, sizeof(char*)); 
        cfg->bypass_from = make_array(p, 1, sizeof(char*));
        cfg->bypass_with = make_array(p, 1, sizeof(char*)); 
        cfg->direct_class = NULL;
        cfg->reject_xrefs = 0;
        cfg->direct_always = 0;
    }
    return (void *) cfg;

}

/* non abbiamo un merge - lasciamo che si overridano */

static void *anoncookies_srvcfg_create(pool *p, server_rec *s) {

    anoncookie_srvconfig *cfg;

    cfg = (anoncookie_srvconfig *) palloc(p, sizeof(anoncookie_srvconfig));
    if (cfg) {
        cfg->classes = make_array(p, 1, sizeof(anoncookie_class));
        cfg->setcookie_name = pstrdup(p, ANONCOOKIES_ID);
        cfg->setcookie_namelen = strlen(ANONCOOKIES_ID);
        cfg->setcookie_path = pstrdup(p, ANONCOOKIES_PATH);
        cfg->setcookie_domain = NULL;
        cfg->cookies_dir = NULL;
        cfg->error_docs = make_array(p, N_ERRORS, sizeof(char *));
       /* NOTA: questo array rimarra' sempre con nelts=0, ma in realta'
          contiene N_ERRORS elementi che setto "a mano" */
    }      
    return (void *) cfg;

}

/* anche qui, niente merge */


/*** FUNZIONI DI SERVIZIO ***/

static anoncookie_class *srv_class_ptr(char *classname, anoncookie_srvconfig *srvcfg) {

/* Restituisce il puntatore alla a_class della classe desiderata, oppure NULL
   se la classe non e' definita. */

    int i = 0;
    anoncookie_class *p = (anoncookie_class *) srvcfg->classes->elts;

    while ((i < srvcfg->classes->nelts) && (strcmp(classname, p[i].class))) {
        i++;
    }
     
    if (i < srvcfg->classes->nelts)
        return &(p[i]);
    else
        return NULL;

}


static int is_required_class(char *classname, anoncookie_dirconfig *dircfg) {

/* Restituisce 1 se la classe e' inclusa nell'elenco di dircfg, altrimenti 0 */

    int i=0;
    char **p = (char **) dircfg->req_classes->elts;

    while ((i < dircfg->req_classes->nelts) && (strcmp(classname, p[i]))) {
        i++;
    }
     
    if (i < dircfg->req_classes->nelts)
        return 1;
    else
        return 0;

}


char *setcookie_val(pool *p, anoncookie_srvconfig *scfg, char *classname, char *cookiename, char *expires) {

/* Genera il valore dell'header Set-Cookie */

    return pstrcat(p, scfg->setcookie_name, "=", classname, ":",
        cookiename, "; path=", scfg->setcookie_path,
        (expires ? "; expires=" : NULL),
        (expires ? expires : NULL),
        (scfg->setcookie_domain ? "; domain=" : NULL),
        (scfg->setcookie_domain ? scfg->setcookie_domain : NULL),
        NULL);

}


static void get_cookiedata_from_filename(anoncookie_data *cookie, char *filename) {

/* Il filename deve essere un filename del giusto formato... niente . e .. ! */
/* Puo' non essere definito il nome (pero' deve essere NULL) e la pool del cookie */

    int i;
    int j = 0;
    char s[MAX_SLEN+1];

    if (!isdigit(filename[0]))       /* e' un cookie normale */
        i = ANONCOOKIE_NAMELEN;
    else if (cookie->name)
        i = strlen(cookie->name);    /* funziona anche con cookie diretti! */
    else {                           /* cookie diretto dal nome sconosciuto */
        i = 0;
        while (filename[i++] != '.');
        i++;
        while (filename[i++] != '.');
        i++;
        while (filename[i++] != '.');
        i++;
        while (filename[i++] != '.');
    }    
    i++;   /* skippo l'ultimo punto */
    
    while (filename[i] != '.') {
        s[j++] = filename[i++];
    }
    s[j] = '\0';
    cookie->expiretime = (time_t) atol(s);
    j = 0;
    i++;
    while (filename[i] != '\0') {
        s[j++] = filename[i++];
    }
    s[j] = '\0';
    cookie->nfiles = atoi(s);
    
}


static anoncookie_data *read_cookie(pool *p, char *dir, char *cookiename) {

/* Esamina i cookie nella directory indicata; se esiste quello richiesto,
   crea una a_data nel pool indicato, la riempie e la restituisce; altrimenti
   restituisce NULL.
   Funziona anche con cookie diretti! */

    int i = 0;
    int not_found = 1;
    int j;
    anoncookie_data *cookie;
    DIR *d;
    struct dirent *dent;
    
    d = opendir(dir);
    if (d == NULL)
        return NULL;
    
    dent = readdir(d);
    while (dent && not_found) {
        while (cookiename[i] == dent->d_name[i])
            i++;
        if (cookiename[i] == '\0')   /* e' uguale fino alla fine del nome */
            not_found = 0;            
        else
            dent = readdir(d);
    }
    
    closedir(d);
    if (not_found)
        return NULL;
        
    cookie = (anoncookie_data *) palloc(p, sizeof(anoncookie_data));
    if (cookie) {
        cookie->pool = p;
        cookie->dir = pstrdup(p, dir);
        cookie->name = pstrdup(p, cookiename);
        cookie->filename = pstrdup(p, dent->d_name);
        get_cookiedata_from_filename(cookie, cookie->filename);
    }
    return cookie;
    
}
              
    
static char *make_new_cookiefilename(anoncookie_data *cookie) {

/* Calcola il nome corretto del cookie e lo alloca in p */

    char s[MAX_SLEN+1];
     
    ap_snprintf(s, MAX_SLEN, "%ld.%d", (long int) cookie->expiretime, cookie->nfiles);
    return pstrcat(cookie->pool, cookie->name, ".", s, NULL);
    
}


static int update_cookie(anoncookie_data *cookie) {

/* Rinomina un cookie da vecchio nome al nome che dovrebbe avere in funzione dei suoi
   attuali dati */
/* Se il cookie ha maxfiles = 1, non e' possibile avere cambiamenti di filename
   del cookie dovuti a richieste parallele; c'e' sempre al piu' una sola
   richiesta che sta maneggiando il cookie. */
/* ...e se anche non ce l'ha, non lo gestisco qui */   

    char *old, *new;
    
    old = pstrcat(cookie->pool, cookie->dir, cookie->filename, NULL);
    if (isdigit(cookie->name[0]) && (!cookie->nfiles)) {   /* e' un cookie diretto da eliminare */
        return unlink(old);
    } else {
        cookie->filename = make_new_cookiefilename(cookie);
        new = pstrcat(cookie->pool, cookie->dir, cookie->filename, NULL);
        return rename(old, new);
    }
    
}    


/* routine al momento inutilizzata */
/* non supporta ancora i cookie diretti */
static int update_multifile_cookie(anoncookie_data *cookie) {

/* Se il cookie ha maxfiles > 1, e' possibile che i download paralleli facciano
   cambiare il suo filename a meta' dell'esecuzione di una singola richiesta.
   In questo caso, prima di fare l'update bisogna cercare nuovamente il cookie. */

    char *old, *new;
    anoncookie_data *ucookie;
    
    old = pstrcat(cookie->pool, cookie->dir, cookie->filename, NULL);
    cookie->filename = make_new_cookiefilename(cookie);
    new = pstrcat(cookie->pool, cookie->dir, cookie->filename, NULL);
    if (rename(old, new)) {
        if (errno == ENOENT) { /* controllo che il nome non sia stato aggiornato */
            ucookie = read_cookie(cookie->pool, cookie->dir, cookie->name);
            if (ucookie) {
                old = pstrcat(cookie->pool, cookie->dir, ucookie->filename, NULL);
                return rename(old, new);
            } else
                return -1;
        } else
            return -1; 
     } else
         return 0;                           
    
}


static int write_cookie(request_rec *r, anoncookie_data *cookie, time_t now) {

    FILE *cfile;
    char *s;
    
    s = pstrcat(cookie->pool, cookie->dir, cookie->filename, NULL);
    cfile = pfopen(cookie->pool, s, "w");
    if (cfile == NULL)
        return 1;
    fprintf(cfile, "%s", ctime(&now));
    s = (char *) get_remote_host(r->connection, r->per_dir_config, REMOTE_NAME);
    if (s && (s[0] != '\0')) 
        fprintf(cfile, "%s\n", s);
    else
        fprintf(cfile, "%s\n", r->connection->remote_ip);
    fprintf(cfile, "%s\n", table_get(r->headers_in, "User-Agent"));
    pfclose(cookie->pool, cfile);
    return 0;
    
}


static void make_new_cookiename(char *s) {

/* Genera un nome di cookie casuale in s (senza controllare i doppioni) */

    int x, j;

    srandom(time(0));
    for(j=0; j<ANONCOOKIE_NAMELEN; j++) {
        do {
            x = random() & 63;
        } while (x > 51);
        if (x < 26)
            s[j] = 'a' + x;
        else
            s[j] = 'A' + x - 26;
    }
    s[j] = '\0';  /* dovrebbe essere inutile */

}


static int is_duplicate_cookie(anoncookie_data *cookie) {

/* Controlla che il nome del cookie non compaia nella directory (quindi
   il controllo va fatto prima di scrivere/aggiornare il cookie!) */

    int i;
    DIR *d;
    struct dirent *dent;
    
    d = opendir(cookie->dir);
    while (dent = readdir(d)) {
        if (dent->d_name[0] != '.') {
            i = 0;
            while (dent->d_name[i] == cookie->name[i])
                i++;
            if (i == ANONCOOKIE_NAMELEN) {
                closedir(d);
                return 1;
            }
        }
    }
    closedir(d);
    return 0;
    
}    


static void create_cookie
   (pool *p, time_t now, anoncookie_class *class, anoncookie_data *cookie) {

    cookie->pool = p;
    cookie->dir = pstrdup(p, class->dir);
    if (class->classtimeout != ANCK_INFINITE_TIMEOUT) 
        cookie->expiretime = now + class->classtimeout;
    else
        cookie->expiretime = ANCK_INFINITE_TIMEOUT;
    cookie->nfiles = 0;
    cookie->name = palloc(p, ANONCOOKIE_NAMELEN+1);
    do {
        make_new_cookiename(cookie->name);
    } while (is_duplicate_cookie(cookie));
    cookie->filename = make_new_cookiefilename(cookie);

}


static void create_direct_cookie
   (pool *p, request_rec *r, anoncookie_class *class, anoncookie_data *cookie) {

    cookie->pool = p;
    cookie->dir = pstrdup(p, class->directdir);
    cookie->expiretime = ANCK_INFINITE_TIMEOUT;
    cookie->nfiles = 0;
    cookie->name = pstrdup(p, r->connection->remote_ip);
    cookie->filename = make_new_cookiefilename(cookie);

}

         
static anoncookie_data *make_cookie(request_rec *r, pool *p, anoncookie_class *class, int *p_error) {

/* Crea, se possibile, un nuovo cookie della classe indicata, altrimenti
   restituisce NULL. La classe e' gia' stata cercata (ossia class != NULL).
   Nel caso restituisca NULL, in *p_error viene messo il codice di errore.
   Alla chiamata, *p_error vale 0 se si vuole creare un cookie normale e 1
   se si vuole un cookie "diretto". */

    int i = 0;
    int direct = *p_error;
    int not_found = 1; 
    int valid = 0;
    anoncookie_data *cookie;
    time_t now = time(NULL);
    DIR *d;
    struct dirent *dent;
    
    cookie = (anoncookie_data *) pcalloc(p, sizeof(anoncookie_data));
    if (cookie == NULL) {
        *p_error = MEMORY_ERROR;
        return NULL;
    }    
    
    d = opendir(class->dir);
    if (d == NULL) {
        *p_error = ANCK_DIR_ERROR;
        return NULL;
    }        
    
    if (class->classtimeout != ANCK_INFINITE_TIMEOUT) {

/* posso sperare di riciclare dei cookie scaduti */
        while (dent = readdir(d)) {
            if (dent->d_name[0] != '.') {
                get_cookiedata_from_filename(cookie, dent->d_name); /* qui non viene allocato il nome */
                if ((!cookie->nfiles) && (now > cookie->expiretime) && (cookie->expiretime != ANCK_INFINITE_TIMEOUT)) {
         /* controllo, cosi' posso comunque avere cookie senza scadenza in classi normali */
                    if (direct)   /* solo qui sotto si alloca il nome del cookie */
                        create_direct_cookie(p, r, class, cookie);
                    else
                        create_cookie(p, now, class, cookie);
                    closedir(d);
                    if (unlink(pstrcat(cookie->pool, class->dir, dent->d_name, NULL))) {  /* cancelliamo e poi chiamiamo write - cosi' update e' limitata ai veri update */
                        *p_error = DIDNT_UPDATE_ANONCOOKIE;
                        return NULL;
                    }    
                    if (write_cookie(r, cookie, now)) {
                        *p_error = ANCK_DIR_ERROR;
                        return NULL;                
                    }    
                    return cookie;          
                } else
                    valid++;
            }        
        }
    } else {

/* i cookie non scadono mai */
        while (dent = readdir(d)) {
            if (dent->d_name[0] != '.')
                valid++;
        }
    }    
    closedir(d);

/* Tiene conto anche dei cookie diretti gia' assegnati */
    d = opendir(class->directdir);
    if (d == NULL) {
        *p_error = ANCK_DIR_ERROR;
        return NULL;
    }        
    while (dent = readdir(d)) {
        if (dent->d_name[0] != '.')
            valid++;
    }
    closedir(d);

    if (valid < class->maxcookies) {
        if (direct)
            create_direct_cookie(p, r, class, cookie);
        else
            create_cookie(p, now, class, cookie);
        if (write_cookie(r, cookie, now)) {
            *p_error = ANCK_DIR_ERROR;
            return NULL;        
        }    
        return cookie;
    }

    *p_error = TOO_MANY_ANONCOOKIES;
    return NULL;

}


static int strstrcasercmp(char *s1, char *s2) {

/* restituisce 0 se la parte finale di s1 coincide con s2, altrimenti 1 */

    int i = strlen(s1);
    int j = strlen(s2);

    if (i < j)       /* s1 e' piu' corta di s2 */
        return 1;
    while ((j >= 0) && (!(s1[i] - s2[j]))) {
        i--;
        j--;
    }
    return (j >= 0);

}


static int match_uagent(request_rec *r, array_header *arr) {

    int i = 0;
    char **p;
    char *uagent;

    uagent = table_get(r->headers_in, "User-Agent");
    p = (char **) arr->elts;
    while (i < arr->nelts) {
        if (!strncasecmp(uagent, *p, strlen(*p)))
            return 1;
        p++;
        i++;
    }
    return 0;

}


static int match_host(request_rec *r, array_header *arr) {

    int i = 0;
    char **p;
    char *host;

    host = (char *) get_remote_host(r->connection, r->per_dir_config, REMOTE_NAME);
    p = (char **) arr->elts;
    while (i < arr->nelts) {
        if (isdigit((*p)[0])) {
            if (!strncasecmp(r->connection->remote_ip, *p, strlen(*p)))
                return 1;
        } else {
            if (!strstrcasercmp(host, *p))
                return 1;
        }
        p++;
        i++;
    }
    return 0;

}


static int must_go_direct(request_rec *r, anoncookie_dirconfig *dcfg) {

    if (dcfg->direct_with->elts) {
        if (match_uagent(r, dcfg->direct_with))
            return 1;
    }

    if (dcfg->direct_from->elts) {
        if (match_host(r, dcfg->direct_from))
            return 1;
    }

    return 0;

}


static int must_bypass(request_rec *r, anoncookie_dirconfig *dcfg) {

    if (dcfg->bypass_with->elts) {
        if (match_uagent(r, dcfg->bypass_with))
            return 1;
    }

    if (dcfg->bypass_from->elts) {
        if (match_host(r, dcfg->bypass_from))
            return 1;
    }

    return 0;

}


anoncookie_reqconfig *get_or_make_reqconfig(request_rec *r) {

    anoncookie_reqconfig *rc;

    rc = (anoncookie_reqconfig *) get_module_config(r->request_config, &anoncookies_module);
    if (rc == NULL) {
        rc = palloc(r->pool, sizeof(anoncookie_reqconfig));
        rc->class = NULL;
        rc->cookie = NULL;
        rc->error = -1;
        set_module_config(r->request_config, &anoncookies_module, (void *) rc);
    }
    return rc;
    
}


static int create_anck_error(request_rec *r, int error) {

    anoncookie_reqconfig *rc;

    rc = get_or_make_reqconfig(r);
    if (rc == NULL)
        return HTTP_INTERNAL_SERVER_ERROR;
    rc->error = error;
    r->handler = ANCK_ERROR_HANDLER_NAME;
    return DECLINED;

}      


static int send_anck_error(request_rec *r, int error) {

    int ret;

    ret = create_anck_error(r, error);
    if (ret == DECLINED)
        return anoncookies_error_handler(r);
    else
        return ret;    

}      


static int redirect_to_uri(request_rec *r, char *uri) {

    request_rec *rr;
    char *new_uri;
    int new_error;

    rr = sub_req_lookup_uri(uri, r);

    if ((rr->status == 200) && (rr->finfo.st_mode != 0)) {  /* l'errordoc e' ok, redirigo su di esso */
        new_uri = escape_uri(r->pool, rr->uri);
        destroy_sub_req(rr);
        internal_redirect(new_uri, r);
        return OK;
    }    
        
    if (is_HTTP_REDIRECT(rr->status)) {   /* l'errordoc ha generato a sua volta un redirect! */
        new_error = rr->status;
        r->notes = overlay_tables(r->pool, r->notes, rr->notes);
        r->headers_out = overlay_tables(r->pool, r->headers_out, rr->headers_out);
        r->err_headers_out = overlay_tables(r->pool, r->err_headers_out, rr->err_headers_out);
        destroy_sub_req(rr);
        return new_error;
    }

    new_error = rr->status;   /* altri errori (es. not found): proseguo */
    if ((new_error == 200) && (!rr->finfo.st_mode))
        new_error = HTTP_NOT_FOUND;
    destroy_sub_req(rr);
    return new_error;
    
}        


static char* reset_class_dir(pool *p, char *dir) {

/* Quando si resetta il server, possono restare dei cookie relativi a scaricamenti interrotti.
   Bisogna resettare il loro numero di file a zero! */

    anoncookie_data *cookie;
    DIR *d;
    struct dirent *dent;
    int i;
    
    cookie = (anoncookie_data *) pcalloc(p, sizeof(anoncookie_data));
    if (cookie == NULL)
        return "Error in allocating memory";
    cookie->pool = p;
    
    d = opendir(dir);
    if (d == NULL)
        return "Error opening class directory in resetting class dir";

    while (dent = readdir(d)) {
        if (dent->d_name[0] != '.') {
            get_cookiedata_from_filename(cookie, dent->d_name);
            if (cookie->nfiles) {
                cookie->nfiles = 0;
                cookie->dir = dir;
                cookie->filename = pstrdup(p, dent->d_name);
                cookie->name = palloc(p, ANONCOOKIE_NAMELEN+1);
                i = 0;
                while (i < ANONCOOKIE_NAMELEN) {
                    cookie->name[i] = dent->d_name[i];
                    i++;
                }
                cookie->name[i] = '\0';     
                if (update_cookie(cookie))
                    return "Couldn't update cookie in resetting class dir";
            }
        }
    }

    closedir(d);
    return NULL;

}


static char* reset_direct_class_dir(pool *p, char *dir) {

    DIR *d;
    struct dirent *dent;
    
    d = opendir(dir);
    if (d == NULL)
        return "Error opening direct class subdirectory in resetting class dir";
    
    while (dent = readdir(d)) {
        if (dent->d_name[0] != '.') {
            if (unlink(pstrcat(p, dir, dent->d_name, NULL)))
                return "Couldn't delete direct cookie in resetting class subdir";
        }
    }

    closedir(d);
    return NULL;

}
               

       
/*** ESECUZIONE DEI COMANDI DI CONFIGURAZIONE ***/

static const char *cmd_setanoncookieclass
   (cmd_parms *cmd, void *dircfg, char *classname, char *classmax) {

    char *rp;
    anoncookie_class *class;
    server_rec *s = cmd->server;
    anoncookie_srvconfig *srvcfg = (anoncookie_srvconfig *) get_module_config(s->module_config, &anoncookies_module);   /* Leggo la per-srv cfg */

    class = (anoncookie_class *) push_array(srvcfg->classes);

    class->class = pstrdup(srvcfg->classes->pool, classname);  /* Alloco la mem. per il nome nella pool e ci copio il nome */
    if (srvcfg->cookies_dir == NULL) {
        class->dir = pstrcat(srvcfg->classes->pool, ANONCOOKIES_DEFAULT_DIR, classname, "/", NULL);
    } else {
        class->dir = pstrcat(srvcfg->classes->pool, srvcfg->cookies_dir, classname, "/", NULL);
    }
    if (mkdir(class->dir, S_IRWXU)) {
        if (errno != EEXIST)
            return "Unable to create the directory for cookie class in SetAnonCookieClass cmd";
        else {
            rp = reset_class_dir(cmd->temp_pool, class->dir);
            if (rp)
                return rp;
        }
    }
    class->directdir = pstrcat(srvcfg->classes->pool, class->dir, ANCK_DIRECT_SUBDIR, NULL);
    if (mkdir(class->directdir, S_IRWXU)) {
        if (errno != EEXIST)
            return "Unable to create the subdirectory for direct cookie class in SetAnonCookieClass cmd";
        else {
            rp = reset_direct_class_dir(cmd->temp_pool, class->directdir);
            if (rp)
                return rp;
        }
    }
    if (chown(class->dir, s->server_uid, s->server_gid))
        return "Unable to chown the cookie-class directory";
    if (chown(class->directdir, s->server_uid, s->server_gid))
        return "Unable to chown the direct-cookie-class subdirectory";
    if (classmax == NULL)
        class->maxcookies = ANCK_INFINITE_COOKIES;
    else
        class->maxcookies = atoi(classmax);
    class->classtimeout = ANCK_DEFAULT_TIMEOUT;
    class->maxfiles = ANCK_DEFAULT_MAXFILES;
    class->ok_doc = NULL;

    return NULL;

}


static const char *cmd_setanoncookiefiles
   (cmd_parms *cmd, void *dircfg, char *maxfiles, char *classname) {

    int n, i;
    anoncookie_class *class;
    server_rec *s = cmd->server;
    anoncookie_srvconfig *srvcfg = (anoncookie_srvconfig *) get_module_config(s->module_config, &anoncookies_module);   /* Leggo la per-srv cfg */

    if (strcasecmp(classname, "all")) {    
        class = srv_class_ptr(classname, srvcfg);
        if (class == NULL)
            return "Error - Undefined class in SetAnonCookieFiles command";
        else {
            n = atoi(maxfiles);
            if (n < 0)
                class->maxfiles = ANCK_INFINITE_MAXFILES;
            else
                class->maxfiles = n;
        }
    } else {
        class = (anoncookie_class *) srvcfg->classes->elts;
        n = atoi(maxfiles);
        if (n < 0) {
            for(i=0; i<srvcfg->classes->nelts; i++) {
                class->maxfiles = ANCK_INFINITE_MAXFILES;
                class++;
            }
        } else {
            for(i=0; i<srvcfg->classes->nelts; i++) {
                class->maxfiles = n;
                class++;
            }
        }
    }                      

    return NULL;

}

          
static const char *cmd_setanoncookieidletime
   (cmd_parms *cmd, void *dircfg, char *idletime, char *classname) {

    int n, i;
    anoncookie_class *class;
    server_rec *s = cmd->server;
    anoncookie_srvconfig *srvcfg = (anoncookie_srvconfig *) get_module_config(s->module_config, &anoncookies_module);   /* Leggo la per-srv cfg */

    if (strcasecmp(classname, "all")) {
        class = srv_class_ptr(classname, srvcfg);
        if (class == NULL)
            return "Error - Undefined class in SetAnonCookieIdletime command";
        else {
            n = atoi(idletime);
            if (n < 0)
                class->classtimeout = ANCK_INFINITE_TIMEOUT;
            else
                class->classtimeout = n;
        }
    } else {
        class = (anoncookie_class *) srvcfg->classes->elts;
        n = atoi(idletime);
        if (n < 0) {
            for(i=0; i<srvcfg->classes->nelts; i++) {
                class->classtimeout = ANCK_INFINITE_TIMEOUT;
                class++;
            }
        } else {
            for(i=0; i<srvcfg->classes->nelts; i++) {
                class->classtimeout = n;
                class++;
            }
        }
    }                      

    return NULL;

}


static const char *cmd_setanoncookieokdoc
   (cmd_parms *cmd, void *dircfg, char *docurl, char *classname) {

    int i;
    anoncookie_class *class;
    server_rec *s = cmd->server;
    anoncookie_srvconfig *srvcfg = (anoncookie_srvconfig *) get_module_config(s->module_config, &anoncookies_module);   /* Leggo la per-srv cfg */
    char *p = pstrdup(srvcfg->classes->pool, docurl);

    if (strcasecmp(classname, "all")) {    
        class = srv_class_ptr(classname, srvcfg);
        if (class == NULL)
            return "Error - Undefined class in SetAnonCookieOkDoc command";
        else 
            class->ok_doc = p;
    } else {
        class = (anoncookie_class *) srvcfg->classes->elts;
        for(i=0; i<srvcfg->classes->nelts; i++) {
            class->ok_doc = p;
            class++;
        }
    }                      

    return NULL;

}


static const char *cmd_setanoncookieerrordoc
   (cmd_parms *cmd, void *dircfg, char *uri, char *errorname) {

    int i;
    char **p;
    char *puri;
    server_rec *s = cmd->server;
    anoncookie_srvconfig *srvcfg = (anoncookie_srvconfig *) get_module_config(s->module_config, &anoncookies_module);   /* Leggo la per-srv cfg */
    p = (char **) srvcfg->error_docs->elts;

    if (!strcasecmp(errorname, "all")) {
        puri = pstrdup(srvcfg->error_docs->pool, uri);
        for(i=0; i<N_ERRORS; i++) {
            *p = puri;
            p++;
        }
        return NULL;
    }    

    if (!strcasecmp(errorname, NEED_ANONCOOKIE_ITEM)) {
        p[NEED_ANONCOOKIE] = pstrdup(srvcfg->error_docs->pool, uri);
        return NULL;
    }    

    if (!strcasecmp(errorname, WRONG_ANONCOOKIE_ITEM)) {
        p[WRONG_ANONCOOKIE] = pstrdup(srvcfg->error_docs->pool, uri);
        return NULL;
    }    

    if (!strcasecmp(errorname, SPURIOUS_ANONCOOKIE_ITEM)) {
        p[SPURIOUS_ANONCOOKIE] = pstrdup(srvcfg->error_docs->pool, uri);
        return NULL;
    }    

    if (!strcasecmp(errorname, UNKNOWN_ANONCOOKIE_ITEM)) {
        p[UNKNOWN_ANONCOOKIE] = pstrdup(srvcfg->error_docs->pool, uri);
        return NULL;
    }    

    if (!strcasecmp(errorname, DEAD_ANONCOOKIE_ITEM)) {
        p[DEAD_ANONCOOKIE] = pstrdup(srvcfg->error_docs->pool, uri);
        return NULL;
    }    

    if (!strcasecmp(errorname, MAXFILES_REACHED_ITEM)) {
        p[MAXFILES_REACHED] = pstrdup(srvcfg->error_docs->pool, uri);
        return NULL;
    }    

    if (!strcasecmp(errorname, TOO_MANY_ANONCOOKIES_ITEM)) {
        p[TOO_MANY_ANONCOOKIES] = pstrdup(srvcfg->error_docs->pool, uri);
        return NULL;
    }    

    if (!strcasecmp(errorname, DIDNT_UPDATE_ANONCOOKIE_ITEM)) {
        p[DIDNT_UPDATE_ANONCOOKIE] = pstrdup(srvcfg->error_docs->pool, uri);
        return NULL;
    }    

    if (!strcasecmp(errorname, MEMORY_ERROR_ITEM)) {
        p[MEMORY_ERROR] = pstrdup(srvcfg->error_docs->pool, uri);
        return NULL;
    }    

    if (!strcasecmp(errorname, ANCK_DIR_ERROR_ITEM)) {
        p[ANCK_DIR_ERROR] = pstrdup(srvcfg->error_docs->pool, uri);
        return NULL;
    }    

    if (!strcasecmp(errorname, REJECT_XREFS_ITEM)) {
        p[REJECT_XREFS] = pstrdup(srvcfg->error_docs->pool, uri);
        return NULL;
    }    

    return "Error - Unknown error type in SetAnonCookieErrorDoc";

}


static const char *cmd_setanoncookiename
   (cmd_parms *cmd, void *dircfg, char *name) {

    anoncookie_srvconfig *srvcfg = (anoncookie_srvconfig *) get_module_config(cmd->server->module_config, &anoncookies_module);   /* Leggo la per-srv cfg */
    srvcfg->setcookie_name = pstrdup(srvcfg->classes->pool, name);
    srvcfg->setcookie_namelen = strlen(name);
    return NULL;    

}    


static const char *cmd_setanoncookiepath
   (cmd_parms *cmd, void *dircfg, char *path) {

    anoncookie_srvconfig *srvcfg = (anoncookie_srvconfig *) get_module_config(cmd->server->module_config, &anoncookies_module);   /* Leggo la per-srv cfg */
    srvcfg->setcookie_path = pstrdup(srvcfg->classes->pool, path);
    return NULL;    

}    


static const char *cmd_setanoncookiedomain
   (cmd_parms *cmd, void *dircfg, char *name) {

    anoncookie_srvconfig *srvcfg = (anoncookie_srvconfig *) get_module_config(cmd->server->module_config, &anoncookies_module);   /* Leggo la per-srv cfg */
    srvcfg->setcookie_domain = pstrdup(srvcfg->classes->pool, name);
    return NULL;    

}    


static const char *cmd_setanoncookiedir
   (cmd_parms *cmd, void *dircfg, char *name) {

    anoncookie_srvconfig *srvcfg = (anoncookie_srvconfig *) get_module_config(cmd->server->module_config, &anoncookies_module);   /* Leggo la per-srv cfg */
    if (name[strlen(name)-1] == '/') {
        srvcfg->cookies_dir = pstrdup(srvcfg->classes->pool, name);
    } else {
        srvcfg->cookies_dir = pstrcat(srvcfg->classes->pool, name, "/", NULL);
    }
    return NULL;    

}    
        

static const char *cmd_requireanoncookies
   (cmd_parms *cmd, void *dircfg, char *classname) {

    char **p;
    array_header *d = ((anoncookie_dirconfig *) dircfg)->req_classes;
    char *s = pstrdup(d->pool, classname);

    if (!strcasecmp(s, "no"))
        d->nelts = 0;
    else {
        p = (char **) push_array(d);
        *p = s;
    }
       
    return NULL;

}


static const char *cmd_rejectexternalrefs
   (cmd_parms *cmd, void *dircfg, char *value) {

    anoncookie_dirconfig *d = (anoncookie_dirconfig *) dircfg;

    if (!strcasecmp(value, "on"))
        d->reject_xrefs = 1;
    else if (!strcasecmp(value, "off"))
        d->reject_xrefs = 0;
    else if (!strcasecmp(value, "explicit"))
        d->reject_xrefs = 2;
    else 
        return "Error - Unknown value in RejectExternalRefs command";
       
    return NULL;

}


static const char *cmd_anoncookiesdirect 
   (cmd_parms *cmd, void *dircfg, char *type, char *value) {

    char **p;
    anoncookie_dirconfig *d = (anoncookie_dirconfig *) dircfg;

    if (!d->direct_class)
        return "Error - Define the AnonCookiesDirectClass before using the AnonCookiesDirect command";

    if (!strcasecmp(type, "always"))
        d->direct_always = 1;
    else if (!strcasecmp(type, "never")) {
        d->direct_always = 0;
        d->direct_with->nelts = 0;
        d->direct_from->nelts = 0;
    } else if (!strcasecmp(type, "with")) {
        d->direct_always = 0;
        p = (char **) push_array(d->direct_with);
        *p = pstrdup(cmd->pool, value);
    } else if (!strcasecmp(type, "from")) {
        d->direct_always = 0;
        p = (char **) push_array(d->direct_from);
        *p = pstrdup(cmd->pool, value);
    } else
        return "Error - Unknown type of permission in AnonCookiesDirect command";

    return NULL;

}


static const char *cmd_anoncookiesdirectclass
   (cmd_parms *cmd, void *dircfg, char *classname) {

    anoncookie_class *class;
    anoncookie_dirconfig *d = (anoncookie_dirconfig *) dircfg;
    anoncookie_srvconfig *srvcfg = (anoncookie_srvconfig *) get_module_config(cmd->server->module_config, &anoncookies_module);   /* Leggo la per-srv cfg */

    class = srv_class_ptr(classname, srvcfg);
    if (!class)
        return "Unknown class in AnonCookiesDirectClass command";
    d->direct_class = class;
    return NULL;

}    


static const char *cmd_anoncookiesbypass 
   (cmd_parms *cmd, void *dircfg, char *type, char *value) {

    char **p;
    anoncookie_dirconfig *d = (anoncookie_dirconfig *) dircfg;

    if (!strcasecmp(type, "with")) {
        p = (char **) push_array(d->bypass_with);
        *p = pstrdup(cmd->pool, value);
    } else if (!strcasecmp(type, "from")) {
        p = (char **) push_array(d->bypass_from);
        *p = pstrdup(cmd->pool, value);
    } else
        return "Error - Unknown type of permission in AnonCookiesBypass command";

    return NULL;

}



/*** HANDLER ***/

static int anoncookies_access_handler(request_rec *r) {

/* Handler per la fase di controllo di accesso - Determina se l'utente ha un anoncookie
   e' se e' sufficiente per accedere al file richiesto */
/* MANCA: in caso di errore, dovrebbe cancellare il cookie (almeno nei casi in cui e' nostro)
   con un "set-cookie" adeguato in err_headers_out */

    anoncookie_dirconfig *dcfg;
    array_header *d;
    anoncookie_srvconfig *scfg;
    char *cookies, *p, *q;
    char *classname;
    char cookiename[ANONCOOKIE_NAMELEN+1];
    anoncookie_reqconfig *rc;
    
    int i, n;
    time_t now;

    if (r->main || r->prev)
        return DECLINED;   /* Questa e' una redirezione/sottorichiesta */

    dcfg = (anoncookie_dirconfig *) get_module_config(r->per_dir_config, &anoncookies_module);
    if (!dcfg)
        return DECLINED;   /* La directory non richiede anoncookies */

    if (dcfg->reject_xrefs) {
        p = table_get(r->headers_in, "Referer");
        if (!p) {   /* non c'e' l'header */
            if (dcfg->reject_xrefs == 1)  /* rifiuta solo se =1 (on) */
                return create_anck_error(r, REJECT_XREFS);
        } else if (strstr(p, "://")) {     /* il referer esiste ed e' assoluto */
            if (!strstr(p, pstrcat(r->pool, "http://", r->server->server_hostname, "/", NULL))) /* ...e non contiene il nostro hostname */
                if (!strstr(p, pstrcat(r->pool, "http://", r->server->server_hostname, ":", NULL))) /* neanche su un'altra porta... */
                   return create_anck_error(r, REJECT_XREFS);   /* rifiuta */
        }
    }

    d = dcfg->req_classes;
    if (!(d && d->nelts))
        return DECLINED;   /* La directory non richiede anoncookies */

/* determino se devo bypassare il resto */

    if (must_bypass(r, dcfg))
        return OK;

/* determino se devo attivare il modo diretto */
    /* controllo se ho gia' un cookie per quell'IP */
    /* se si', lo leggo e proseguo dritto */
    /* se no, lo creo e proseguo dritto */
    /* o meglio, taglio tutto fino al read_cookie */
/* il modo diretto bypassa il controllo di classe: posso assegnare cookie
   di una classe non richiesta */

    if (dcfg->direct_always || must_go_direct(r, dcfg)) {
        
        /* Direct mode */

        rc = get_or_make_reqconfig(r);
        rc->class = dcfg->direct_class;
        rc->cookie = read_cookie(r->pool, dcfg->direct_class->directdir, r->connection->remote_ip);
        if (!rc->cookie) {
            n = 1;   /* forza la creazione di un cookie diretto */
            rc->cookie = make_cookie(r, r->pool, dcfg->direct_class, &n);
            if (rc->cookie == NULL)
                return create_anck_error(r, n);
        }
        scfg = (anoncookie_srvconfig *) get_module_config(r->server->module_config, &anoncookies_module);
        if (!scfg)
            return create_anck_error(r, MEMORY_ERROR);

    } else {

        /* Modo normale */

        cookies = table_get(r->headers_in, "Cookie");
        if (cookies == NULL)
            return create_anck_error(r, NEED_ANONCOOKIE);  /* L'utente non ha alcun anoncookie */

        scfg = (anoncookie_srvconfig *) get_module_config(r->server->module_config, &anoncookies_module);
        if (!scfg)
            return create_anck_error(r, MEMORY_ERROR);
        n = 1;
        p = strstr(cookies, scfg->setcookie_name);
        while (p && n) {
            if (p == cookies) {
                n = (*(p + scfg->setcookie_namelen) != '=');
            } else {
                n = ((*(p + scfg->setcookie_namelen) != '=') || (*(p-1) != ' '));
            }
            if (n)
                p = strstr(p, scfg->setcookie_name);
        }    

        if (p == NULL)
            return create_anck_error(r, NEED_ANONCOOKIE);  /* L'utente non ha alcun anoncookie */
        p += scfg->setcookie_namelen;
        p++;
        q = p;
        while (((*q) != ':') && ((*q) != '\0'))
            q++;
        n = (int) (q-p);
        classname = palloc(r->pool, n+1);
        i = 0;
        while (i < n)
            classname[i++] = *(p++);
        classname[i] = '\0';
        rc = get_or_make_reqconfig(r);
        rc->class = srv_class_ptr(classname, scfg);

        if (rc->class == NULL)
            return create_anck_error(r, UNKNOWN_ANONCOOKIE);    /* Il cookie e' di una classe sconosciuta (falso? vecchio?) */
        if (!is_required_class(classname, dcfg))
            return create_anck_error(r, WRONG_ANONCOOKIE);      /* La classe esiste, ma non e' di quelle richieste per questa dir */
        p++;
        i = 0;
        while (i < ANONCOOKIE_NAMELEN)
            cookiename[i++] = *(p++);
        cookiename[i] = '\0';
        rc->cookie = read_cookie(r->pool, rc->class->dir, cookiename);
        if (rc->cookie == NULL) {
            table_set(r->err_headers_out, "Set-Cookie", setcookie_val(r->pool, scfg, classname, cookiename, THE_PAST));
            return create_anck_error(r, SPURIOUS_ANONCOOKIE);   /* La classe esiste, ma il cookie no (gia' riciclato? falso?) */
        }
    }

    if ((rc->class->maxfiles != ANCK_INFINITE_MAXFILES) && (rc->cookie->nfiles >= rc->class->maxfiles))
        return create_anck_error(r, MAXFILES_REACHED);      /* L'utente ha gia' troppi scaricamenti in corso */
    if (!(rc->cookie->nfiles)) {
        now = time(NULL);
        if ((now > rc->cookie->expiretime) && (rc->cookie->expiretime != ANCK_INFINITE_TIMEOUT)) {
            if (unlink(pstrcat(rc->cookie->pool, rc->cookie->dir, rc->cookie->filename, NULL)))
                return create_anck_error(r, ANCK_DIR_ERROR);
            table_set(r->err_headers_out, "Set-Cookie", setcookie_val(r->pool, scfg, classname, cookiename, THE_PAST));
            return create_anck_error(r, DEAD_ANONCOOKIE);   /* Il cookie e' morto */
        }
    }

    (rc->cookie->nfiles)++;

    if (update_cookie(rc->cookie))
        return create_anck_error(r, DIDNT_UPDATE_ANONCOOKIE);
    return OK;

}


/* Rimozione cookie morti: la cosa migliore sarebbe farlo una volta ogni tanto...
   Alternativa: riciclo i cookie morti quando devo assegnarne uno (cambiando il nome)!! in questo
   modo l'array diventa grande il massimo, ma tanto piu' di N non ne assegno :-)
   Nota che un cookie e' espirato se time > expiretime _e_ nfiles = 0!
   ok, va bene il riciclaggio
 */


static int anoncookies_endreq_handler(request_rec *r) {

/* Handler per la fase di log - Non logga alcunche', ma essendo eseguito a fine richiesta
   serve ad aggiornare il conteggio degli scaricamenti in corso */

    anoncookie_reqconfig *rc;
    time_t now;

/* in questo caso non chiamo get_or_make, perche' voglio vedere se c'e' o no... 
   se non c'e' dentro un cookie, non mi interessa */
    rc = (anoncookie_reqconfig *) get_module_config(r->request_config, &anoncookies_module);
    if (!rc)
        return DECLINED;  /* l'access handler ha declinato: non era una dir. protetta, era una sottorichiesta, ecc... */
    if ((rc->error >= 0) || (!rc->cookie))
        return DECLINED;  /* in caso di errore, non devo aggiornare cookie! */   
        
/* Se maxfiles > 1, rileggo il cookie perche' il suo stato puo' essere stato
   aggiornato da una richiesta parallela */
    if (rc->class->maxfiles > 1) {    
        rc->cookie = read_cookie(rc->cookie->pool, rc->cookie->dir, rc->cookie->name);
        if (!rc->cookie) /* c'e' stato un errore, ma non posso fare piu' nulla */
            return DECLINED;
    }

    if ((rc->cookie->nfiles <= 1) && (rc->cookie->expiretime != ANCK_INFINITE_TIMEOUT)) {
        now = time(NULL);
        now += rc->class->classtimeout;
        rc->cookie->expiretime = now;
    }

    (rc->cookie->nfiles)--;
    update_cookie(rc->cookie);   /* non posso piu' segnalare errori */
    return OK;

}


static int anoncookies_get_handler(request_rec *r) {

/* Handler per le richieste di cookie - Non viene mai richiamato, se non esplicitamente dalla
   configurazione di Apache, con un Add/SetHandler. Il tipo del cookie desiderato deve essere
   l'ultima parte (dopo l'ultima / ) dell'URI richiesto. */

    char *classname;
    int i, l, n;
    anoncookie_srvconfig *scfg;
    anoncookie_class *class;
    anoncookie_data *cookie;

    l = strlen(r->uri);
    if (r->uri[l-1] == '/')
        l--;      /* Elimina una eventuale / finale */
    i = l;
    i--;
    while (r->uri[i] != '/')
        i--;
    n = l-i;
    classname = palloc(r->pool, n);
    n--;
    l = 0;
    while (l < n)
        classname[l++] = r->uri[++i];
    classname[l] = '\0';
    scfg = (anoncookie_srvconfig *) get_module_config((r->server)->module_config, &anoncookies_module);
    class = srv_class_ptr(classname, scfg);
    if (class == NULL)
        return send_anck_error(r, UNKNOWN_ANONCOOKIE);
    /* Il cookie richiesto e' di una classe sconosciuta */
    n = 0;    /* chiede un cookie normale */
    cookie = make_cookie(r, r->pool, class, &n);
    if (cookie == NULL)
        return send_anck_error(r, n);
  /* E' gia' stato raggiunto il numero massimo di cookie della classe richiesta */

  /* il cookie va in err_headers altrimenti non si propaga all'eventuale redirect */  
    table_set(r->err_headers_out, "Set-Cookie", setcookie_val(r->pool, scfg, classname, cookie->name, THE_FUTURE));   
    table_set(r->err_headers_out, "Pragma", "no-cache");   /* Don't cache responses */

    n = 0;
    if (class->ok_doc) {
         n = redirect_to_uri(r, class->ok_doc);
         if (n == OK)
             return OK;
    }
            
    r->content_type = "text/html";
    soft_timeout("send", r);
    send_http_header(r);
            
    if (!r->header_only) {
        rputs("<HTML>\n<HEAD><TITLE>AnonCookie given</TITLE></HEAD>\n", r);
        rprintf(r, "<BODY><H3>Anoncookie of class %s given</H3><HR>", classname);
        rprintf(r, "You have received a cookie of class %s. Now you can access", classname);
        rputs(" all the URLs which require that class of cookies.", r);
        if (n) {
            rputs("\n<P>Additionally, the webmaster asked us to send you the document at the URL ", r);
            rprintf(r, "%s, but we found an HTTP error code of %d. Kick the webmaster.", class->ok_doc, n);
        }
        rputs("\n</BODY></HTML>", r);
    }    

    kill_timeout(r);
    return OK;

}


static int anoncookies_error_handler(request_rec *r) {

/* Handler per gli errori di generazione e gestione dei cookie. Non viene
   mai invocato, se non internamente settando il r->handler. */
   
    char **p_errordoc;
    int new_error = 0;
    anoncookie_srvconfig *scfg;
    anoncookie_reqconfig *rc;

    scfg = (anoncookie_srvconfig *) get_module_config(r->server->module_config, &anoncookies_module);
    rc = (anoncookie_reqconfig *) get_module_config(r->request_config, &anoncookies_module);
    
    if ((!rc) || (rc->error < 0) || (rc->error >= N_ERRORS)) 
        return DECLINED;

    table_set(r->err_headers_out, "Pragma", "no-cache");   /* Don't cache errors */

    p_errordoc = (char **) scfg->error_docs->elts;
    p_errordoc += rc->error;
    if (*p_errordoc) {    /* E' definito un errordoc */
        new_error = redirect_to_uri(r, *p_errordoc);
        if (new_error == OK)
            return OK;
    }

    r->content_type = "text/html";
    soft_timeout("send", r);
    send_http_header(r);
            
    if (!r->header_only) {
        rputs("<HTML>\n<HEAD><TITLE>AnonCookies Error</TITLE></HEAD>\n", r);
        switch (rc->error) {

            case NEED_ANONCOOKIE:
                rprintf(r, "<BODY><H3>'%s' Anoncookies Error</H3><HR>\n", NEED_ANONCOOKIE_ITEM);
                rputs("You need an anonymous cookie of the correct class to access this URL.", r);
                break;

            case UNKNOWN_ANONCOOKIE:
                rprintf(r, "<BODY><H3>'%s' Anoncookies Error</H3><HR>\n", UNKNOWN_ANONCOOKIE_ITEM);
                rputs("The class of your (requested) anonymous cookie is unknown. Gee, you really messed with it.", r);
                break;
                
            case WRONG_ANONCOOKIE:
                rprintf(r, "<BODY><H3>'%s' Anoncookies Error</H3><HR>\n", WRONG_ANONCOOKIE_ITEM);
                rputs("You have an anonymous cookie of the wrong class. You need a cookie of the correct class to access this URL.", r);
                break;

            case SPURIOUS_ANONCOOKIE:
                rprintf(r, "<BODY><H3>'%s' Anoncookies Error</H3><HR>\n", SPURIOUS_ANONCOOKIE_ITEM);
                rputs("You have an anonymous cookie of the correct class, but it is dead (maybe old, or maybe you ", r);
                rputs("stayed idle too long and it expired). You need a valid cookie of the correct class to access this URL.", r);
                break;

            case DEAD_ANONCOOKIE:
                rprintf(r, "<BODY><H3>'%s' Anoncookies Error</H3><HR>\n", DEAD_ANONCOOKIE_ITEM);
                rputs("You have an anonymous cookie of the correct class, but you stayed idle too long and it expired. You need a valid cookie of the correct class to access this URL.", r);
                break;

            case MAXFILES_REACHED:
                rprintf(r, "<BODY><H3>'%s' Anoncookies Error</H3><HR>\n", MAXFILES_REACHED_ITEM);
                rputs("You are already downloading the maximum number of files allowed by your cookie. Wait until you finish downloading the other files.", r);
                break;

            case TOO_MANY_ANONCOOKIES:
                rprintf(r, "<BODY><H3>'%s' Anoncookies Error</H3><HR>\n", TOO_MANY_ANONCOOKIES_ITEM);
                rputs("We already distributed all the cookies of this class. Try again until someone else releases a cookie.", r);
                break;

            case DIDNT_UPDATE_ANONCOOKIE:
                rprintf(r, "<BODY><H3>'%s' Anoncookies Error</H3><HR>\n", DIDNT_UPDATE_ANONCOOKIE_ITEM);
                rputs("Uh-oh. Our server couldn't update the cookie. Maybe someone grabbed the cookie we were about to send you...", r);
                break;

            case MEMORY_ERROR:
                rprintf(r, "<BODY><H3>'%s' Anoncookies Error</H3><HR>\n", MEMORY_ERROR_ITEM);
                rputs("Uh-oh. We ran out of memory, or had problems allocating things. Gosh.", r);
                break;

            case ANCK_DIR_ERROR:
                rprintf(r, "<BODY><H3>'%s' Anoncookies Error</H3><HR>\n", ANCK_DIR_ERROR_ITEM);
                rputs("Uh-oh. We couldn't read/write in the cookies dir. Kick the webmaster.", r);
                break;

            case REJECT_XREFS:
                rprintf(r, "<BODY><H3>'%s' Anoncookies Error</H3><HR>\n", REJECT_XREFS_ITEM);
                rputs("You tried to access this URL from a link on another server. We don't accept external links to this URL - visit our site, instead.", r);
                break;

            default:
                rputs("<BODY><H3>AAAAAAAGH!</H3><HR>Anoncookie error of an unknown type. Gosh.", r);
        
        }
        if (new_error) {
            rputs("\n<P>Additionally, the webmaster asked us to send you the error document at the URL ", r);
            rprintf(r, "%s, but we found an HTTP error code of %d. Kick the webmaster.", *p_errordoc, new_error);
        }
        rputs("\n</BODY></HTML>", r);
    }    

    kill_timeout(r);
    return OK;

}                                    
                              


/*** TABELLE DI INTERFACCIA DEL MODULO ***/

/* Comandi di configurazione */

command_rec anoncookies_commands[] = {

    { "SetAnonCookieClass", cmd_setanoncookieclass, NULL, RSRC_CONF, TAKE12,
          "Defines a class of anonymous cookies and their max number" },
    { "SetAnonCookieFiles", cmd_setanoncookiefiles, NULL, RSRC_CONF, ITERATE2,
          "Sets the max number of parallel downloads for specified class(es)" },
    { "SetAnonCookieIdletime", cmd_setanoncookieidletime, NULL, RSRC_CONF, ITERATE2,
          "Sets the max idle time for specified class(es)" },
    { "SetAnonCookieName", cmd_setanoncookiename, NULL, RSRC_CONF, TAKE1,
          "Sets the name of the anoncookies sent to the user-agents" },
    { "SetAnonCookiePath", cmd_setanoncookiepath, NULL, RSRC_CONF, TAKE1,
          "Sets the path of the anoncookies sent to the user-agents" },
    { "SetAnonCookieDomain", cmd_setanoncookiedomain, NULL, RSRC_CONF, TAKE1,
          "Sets the domain of the anoncookies sent to the user-agents" },
    { "SetAnonCookieDir", cmd_setanoncookiedir, NULL, RSRC_CONF, TAKE1,
          "Sets the dir in which anoncookies will be stored" },
    { "SetAnonCookieErrorDoc", cmd_setanoncookieerrordoc, NULL, RSRC_CONF, ITERATE2,
          "Sets the error document URI for the specified error(s)" },
    { "SetAnonCookieOkDoc", cmd_setanoncookieokdoc, NULL, RSRC_CONF, ITERATE2,
          "Sets the document URI to be sent when cookies of the specified class(es) are assigned" },
    { "RequireAnonCookies", cmd_requireanoncookies, NULL, OR_LIMIT|OR_AUTHCFG, ITERATE,
          "Sets the cookie class(es) required to access the directory" },
    { "RejectExternalRefs", cmd_rejectexternalrefs, NULL, OR_LIMIT|OR_AUTHCFG, TAKE1,
          "Sets whether external links to this directory are allowed" },
    { "AnonCookiesDirect", cmd_anoncookiesdirect, NULL, OR_LIMIT|OR_AUTHCFG, ITERATE2,
          "Sets the 'direct' cookie mode for specified IPs/browsers" },
    { "AnonCookiesDirectClass", cmd_anoncookiesdirectclass, NULL, OR_LIMIT|OR_AUTHCFG, TAKE1,
          "Sets the class of cookies given in direct mode" },
    { "AnonCookiesBypass", cmd_anoncookiesbypass, NULL, OR_LIMIT|OR_AUTHCFG, ITERATE2,
          "Lets specified IPs/browsers bypass anonymous cookie checks" },
    { NULL }

};


/* Handler direttamente accessibili */

handler_rec anoncookies_handlers[] = {
    { "get-anoncookie", anoncookies_get_handler },
    { ANCK_ERROR_HANDLER_NAME, anoncookies_error_handler },
    { NULL }
};          


/* Definizione del modulo */

module anoncookies_module = {
    STANDARD_MODULE_STUFF,
#ifdef DEBUG
    anoncookies_init,            /* initializer */
#else
    NULL,                        /* initializer */
#endif
    anoncookies_dircfg_create,   /* dir config creator */
    NULL,                        /* dir merger --- default is to override */
    anoncookies_srvcfg_create,   /* server config */
    NULL,                        /* merge server config */
    anoncookies_commands,        /* command table */
    anoncookies_handlers,        /* handlers */
    NULL,                        /* filename translation */
    NULL,                        /* check_user_id */
    NULL,                        /* check auth */
    anoncookies_access_handler,  /* check access */
    NULL,                        /* type_checker */
    NULL,                        /* fixups */
    anoncookies_endreq_handler,  /* logger */
    NULL                         /* header parser */
};   
